from univariate.anomaly.detector.anomaly_detector import AnomalyDetector
from univariate.anomaly.detector.iqr_outlier_detector import IQROutlierDetector
from univariate.anomaly.detector.residual_anomaly_detector import ResidualAnomalyDetector



"""

"""
import pendulum
from pyspark.sql import SparkSession
from pyspark.sql.types import StructType, StructField, LongType
from univariate.sampling.utils import freq_to_period_map
from datetime import datetime, timedelta
from typing import List, Tuple
from pendulum import period


def construct_range_array(sampling_period: int, first_start: int, last_start: int, by_freq: bool) -> Tuple[List[int], List[int]]:
    """
    Construct range tuple that contains each ("start", "end") array representing (2, n) shape
    Note: Each range is half-closed interval
    :param sampling_period:
    :param first_start:
    :param last_start:
    :param by_freq:
    :return:
    """
    if by_freq:
        start = pendulum.from_timestamp(int(first_start / 1000))
        end = pendulum.from_timestamp(int(last_start / 1000))
        period = pendulum.period(start, end)

        #unit_dict = {'A': 'years', 'M': 'months', 'W': 'weeks', 'D': 'days', 'H': 'hours', 'Q': 'months'}
        try:
            freq = list(freq_to_period_map.keys())[list(freq_to_period_map.values()).index(sampling_period)]
        except Exception as e:
            raise ValueError(f"Can't convert sampling period {sampling_period} to freq")
        if freq == "A":
            start_series = list(map(lambda x: int(x.timestamp() * 1000), period.range("years")))
            end_series = list(map(lambda x: int(x.add(years=1).timestamp() * 1000), period.range("years")))
        elif freq == "M":
            start_series = list(map(lambda x: int(x.timestamp() * 1000), period.range("months")))
            end_series = list(map(lambda x: int(x.add(months=1).timestamp() * 1000), period.range("months")))
        elif freq == "W":
            start_series = list(map(lambda x: int(x.timestamp() * 1000), period.range("weeks")))
            end_series = list(map(lambda x: int(x.add(weeks=1).timestamp() * 1000), period.range("weeks")))
        elif freq == "D":
            start_series = list(map(lambda x: int(x.timestamp() * 1000), period.range("days")))
            end_series = list(map(lambda x: int(x.add(days=1).timestamp() * 1000), period.range("days")))
        elif freq == "H":
            start_series = list(map(lambda x: int(x.timestamp() * 1000), period.range("hours")))
            end_series = list(map(lambda x: int(x.add(hours=1).timestamp() * 1000), period.range("hours")))
        elif freq == 'Q':
            start_series = list(map(lambda x: int(x.timestamp() * 1000), period.range("months", 3)))
            end_series = list(map(lambda x: int(x.add(months=3).timestamp() * 1000), period.range("months", 3)))
        else:
            raise ValueError(f"Freq {freq} has no rule")
    else:
        start_series = list(range(first_start, last_start + 1, sampling_period))
        end_series = list(range(first_start + sampling_period, last_start + sampling_period + 1, sampling_period))

    return start_series, end_series


def find_start_timestamp_by_freq(observed_at: int, sampling_period: int) -> int:
    """
    Find start timestamp of range which observed_at arguments are belonged
    :param observed_at:
    :param sampling_period:
    :return:
    """
    start_timestamp: int
    try:
        freq = list(freq_to_period_map.keys())[list(freq_to_period_map.values()).index(sampling_period)]
    except Exception as e:
        raise ValueError(f"can't convert sampling period {sampling_period} to freq")
    observed_date = datetime.fromtimestamp(observed_at / 1000)
    if freq == 'A':  # yearly
        start_timestamp = datetime(year=observed_date.year, month=1, day=1).timestamp() * 1000
    elif freq == 'Q':
        start_timestamp = datetime(year=observed_date.year,
                                   month=3 * ((observed_date.month - 1) // 3) + 1,
                                   day=1).timestamp() * 1000
    elif freq == 'M':
        start_timestamp = datetime(year=observed_date.year, month=observed_date.month, day=1).timestamp() * 1000
    elif freq == 'W':
        start_of_week = observed_date - timedelta(days=observed_date.weekday())
        start_timestamp = datetime(year=start_of_week.year, month=start_of_week.month, day=start_of_week.day).timestamp() * 1000
    elif freq == 'D':
        start_timestamp = datetime(year=observed_date.year, month=observed_date.month, day=observed_date.day).timestamp() * 1000
    elif freq == 'H':
        start_timestamp = datetime(year=observed_date.year, month=observed_date.month, day=observed_date.day, hour=observed_date.hour).timestamp() * 1000
    else:
        raise ValueError(f"freq {freq} has no converting rule")
    return start_timestamp



"""

"""

from univariate.missing_value.handler.strategy import MissingValueHandleStrategy, MissingValueHandleStrategyType
from pyspark.sql import DataFrame
import pyspark.sql.functions as F
from univariate.analyzer import AnalysisReport
from univariate.analyzer import DescriptiveStatAnalyzer
from typing import Optional
from pyspark.sql.window import Window
from scipy.interpolate import lagrange
import numpy as np
from pyspark.sql.types import ArrayType, StructType, StructField, LongType, FloatType


__all__ = ["LagrangeInterpolation"]


class LagrangeInterpolation(MissingValueHandleStrategy):
    """ """

    def handle(
            self, missed_ts: DataFrame, time_col_name: str, data_col_name, handled_col_name: str, k: Optional[int] = 3
    ) -> DataFrame:
        """
        todo: explain "pointwise lagrange"
        :param missed_ts:
        :param time_col_name:
        :param data_col_name:
        :param handled_col_name:
        :param k:
        :return:
        """
        # todo: documentation

        # todo: adjusted k, parameterized
        separated_df = missed_ts.withColumn('not_null_timestamp', F.when(F.col(data_col_name).isNotNull(), F.col(time_col_name)))
        not_null_df = missed_ts.filter(F.col(data_col_name).isNotNull())

        left_window = Window.orderBy(time_col_name).rowsBetween(-k + 1, Window.currentRow)
        right_window = Window.orderBy(time_col_name).rowsBetween(Window.currentRow, k - 1)
        left_stretched_df = not_null_df.select(
            F.col(time_col_name).alias('center'),
            F.collect_list(time_col_name).over(left_window).alias("left_timestamps"),
            F.collect_list(data_col_name).over(left_window).alias("left_values"),
        )
        right_stretched_df = not_null_df.select(
            F.col(time_col_name).alias('center'),
            F.collect_list(time_col_name).over(right_window).alias("right_timestamps"),
            F.collect_list(data_col_name).over(right_window).alias("right_values"),
        )

        # create relative references to the start value (last value not missing)
        w_start = Window.orderBy(time_col_name).rowsBetween(Window.unboundedPreceding, -1)
        w_end = Window.orderBy(time_col_name).rowsBetween(0, Window.unboundedFollowing)
        bounded_df = separated_df.withColumn('left_not_null_timestamp', F.when(F.col(data_col_name).isNull(),
                                                                               F.last('not_null_timestamp', True).over(
                                                                                   w_start))) \
            .withColumn('right_not_null_timestamp',
                        F.when(F.col(data_col_name).isNull(), F.first('not_null_timestamp', True).over(w_end)))
        joined_df = bounded_df.join(left_stretched_df,
                                    bounded_df['left_not_null_timestamp'] == left_stretched_df['center'], "left") \
            .join(right_stretched_df, bounded_df['right_not_null_timestamp'] == right_stretched_df['center'],
                  "left")

        select_k_nearest_udf = F.udf(select_k_nearest, ArrayType(StructType([
            StructField("timestamp", LongType()),
            StructField("value", FloatType()),
        ])))

        nearest_df = joined_df.withColumn("nearest", F.when(F.col(data_col_name).isNull(),
                                                            select_k_nearest_udf(F.col("event_time"),
                                                                                 F.col("left_values"),
                                                                                 F.col("left_timestamps"),
                                                                                 F.col("right_values"),
                                                                                 F.col("right_timestamps"),
                                                                                 F.lit(k))))

        interpolate_with_lagrange_udf = F.udf(interpolate_with_lagrange, FloatType())
        filled_df = nearest_df.withColumn(handled_col_name, F.when(F.col("input_data").isNull(),
                                                           interpolate_with_lagrange_udf(F.col("nearest"), F.col(
                                                               "event_time"))).otherwise(F.col("input_data"))).drop('not_null_timestamp', 'left_not_null_timestamp', 'right_not_null_timestamp', 'center', 'left_timestamps', 'left_values', 'right_timestamps', 'right_values', 'nearest').sort(time_col_name)
        return filled_df


def select_k_nearest(timestamp, left_values, left_timestamps, right_values, right_timestamps, k=10):
    cnt = 0
    left = 0
    right = 0
    left_size = len(left_values) if left_values is not None else 0
    right_size = len(right_values) if right_values is not None else 0

    nearest = []
    while cnt < k and (left < left_size or right < right_size):
        if right >= right_size:
            left = left + 1
            nearest.append([left_timestamps[left_size - left], left_values[left_size - left]])
        elif left >= left_size:
            nearest.append([right_timestamps[right], right_values[right]])
            right = right + 1
        elif (timestamp - left_timestamps[left_size - left - 1]) <= (right_timestamps[right] - timestamp):
            left = left + 1
            nearest.append([left_timestamps[left_size - left], left_values[left_size - left]])
        else:
            nearest.append([right_timestamps[right], right_values[right]])
            right = right + 1
        cnt = cnt + 1

    return nearest


def interpolate_with_lagrange(observations, timestamp):
    if observations is None:
        return None
    x = np.array([observation[0] for observation in observations])
    xm = np.mean(x)
    xscale = np.std(x)

    y = np.array([observation[1] for observation in observations])
    ym = np.mean(y)
    yscale = np.std(y)

    x = (x - xm) / xscale
    y = (y - ym) / yscale

    poly = lagrange(x, y)
    y_hat = poly((timestamp - xm) / xscale)

    estimate = (y_hat * yscale + ym).item()
    return estimate



"""

"""

from univariate.sampling.strategy import SamplingStrategy
from pyspark.sql import DataFrame
from typing import Optional, List


class LinearUpsampling(SamplingStrategy):
    """

    """
    @staticmethod
    def sample(ts: DataFrame, time_col_name: str, data_col_name: str, period: int, sampling_period: int, by_freq: bool, count_flag_cols: Optional[List[str]] = None) -> DataFrame:
        """

        :param ts:
        :param time_col_name:
        :param data_col_name:
        :param period:
        :param sampling_period:
        :param by_freq:
        :return:
        """




"""

"""

from pyspark.sql import DataFrame
from univariate.analyzer import AnalysisReport
from univariate.missing_value.handler.strategy import MissingValueHandleStrategy, MissingValueHandleStrategyType
import importlib
import plotly.graph_objects as go
from plotly.graph_objects import Figure
import plotly.express as px
from typing import Optional


class MissingValueHandler:
    """
    It can handle missing value with some strategies, such as imputation, deletion, filling constant, interpolation, and so on
    """
    def __init__(self, handler_type: MissingValueHandleStrategyType):
        concrete_cls = getattr(
            importlib.import_module("univariate.missing_value.handler.strategy"),
            handler_type.value
        )
        self.__handler_type: MissingValueHandleStrategyType = handler_type
        self.__handler: MissingValueHandleStrategy = concrete_cls()


    def handle_missing_value(self, missed_ts: DataFrame, time_col_name: str, data_col_name: str, handled_col_name: Optional[str] = None, **kwargs) -> AnalysisReport:
        """

        :param missed_ts:
        :param time_col_name:
        :param data_col_name:
        :param handled_col_name: # todo: explain
        :param kwargs:
        :return:
        """
        # todo: decision logic decoupling, decision rule-engine?
        if handled_col_name is None:
            handled_col_name = data_col_name + "_mv_handled"
        filled_ts = self.__handler.handle(missed_ts, time_col_name, data_col_name, handled_col_name, **kwargs)
        report = self.__build_report(missed_ts, filled_ts, time_col_name, data_col_name, handled_col_name)
        return report


    def __build_report(self, missed_ts: DataFrame, filled_ts: DataFrame, time_col_name: str, data_col_name: str, handled_col_name: str) -> AnalysisReport:
        """

        :param missed_ts:
        :param filled_ts:
        :param time_col_name:
        :param data_col_name:
        :return:
        """
        # todo: refactor report system
        report = AnalysisReport()
        report.parameters['strategy'] = self.__handler_type.value
        # report.parameters['mean'] = mean  # todo: build logic with specific strategy
        # report.parameters['ad-hoc'] = ad_hoc

        report.parameters['missed_ts'] = missed_ts
        report.parameters['filled_ts'] = filled_ts

        report.plots['disconnected_plot'] = self.__build_disconnected_plot(filled_ts, time_col_name, data_col_name)
        report.plots['connected_plot'] = self.__build_connected_plot(filled_ts, time_col_name, data_col_name, handled_col_name)

        return report


    def __build_disconnected_plot(self, filled_ts: DataFrame, time_col_name: str, data_col_name: str) -> Figure:
        """

        :param filled_ts:
        :param time_col_name:
        :param data_col_name:
        :return:
        """
        pd_df = filled_ts.select(time_col_name, data_col_name).sort(time_col_name).toPandas()
        fig = px.line(pd_df, x=time_col_name, y=data_col_name)
        return fig


    def __build_connected_plot(self, filled_ts: DataFrame, time_col_name: str, data_col_name: str, handled_col_name: str) -> Figure:
        """

        :param filled_ts:
        :param time_col_name:
        :param data_col_name:
        :param handled_col_name:
        :return:
        """
        pd_df = filled_ts.select(time_col_name, data_col_name, handled_col_name).sort(time_col_name).toPandas()
        fig = px.line(pd_df, x=time_col_name, y=handled_col_name)
        fig.add_trace(go.Scatter(x=pd_df[pd_df[data_col_name].isnull()][time_col_name],
                                 y=pd_df[pd_df[data_col_name].isnull()][handled_col_name],
                                 mode='markers',
                                 name='replaced value'))
        return fig



"""

"""

from abc import ABCMeta
from pyspark.sql import DataFrame
from univariate.analyzer import AnalysisReport
from typing import Optional

class AnomalyDetector(metaclass=ABCMeta):
    """

    """
    @staticmethod
    def detect(ts: DataFrame, **kwargs) -> AnalysisReport:
        """

        :param ts:
        :param kwargs:
        :return:
        """
        pass



"""

"""

from univariate.missing_value.strategy.unmarked_detection import (
    UnmarkedMissingValueDetectionStrategy,
)
from pyspark.sql import DataFrame
from univariate.analyzer import AnalysisReport


class DoNotDetect(UnmarkedMissingValueDetectionStrategy):
    """
    Do nothing
    """

    @staticmethod
    def detect(
        ts: DataFrame,
        time_col_name: str,
        data_col_name: str,
        period: int
    ) -> DataFrame:
        return ts



"""

"""

from univariate.missing_value.handler.strategy import MissingValueHandleStrategy, MissingValueHandleStrategyType
from pyspark.sql import DataFrame
import pyspark.sql.functions as F
from pyspark.sql.window import Window


class BOCFImputation(MissingValueHandleStrategy):
    """ """

    def handle(
        self, missed_ts: DataFrame, time_col_name: str, data_col_name, handled_col_name: str) -> DataFrame:
        """

        :param missed_ts:
        :param time_col_name:
        :param data_col_name:
        :param handled_col_name:
        :return:
        """
        window = Window.orderBy(time_col_name).rowsBetween(Window.currentRow, Window.unboundedFollowing)
        filled_ts = missed_ts.withColumn(handled_col_name, F.first(data_col_name, ignorenulls=True).over(window))
        return filled_ts



"""

"""

from pyspark.sql import DataFrame
from univariate.analyzer import AnalysisReport
from univariate.anomaly.detector import AnomalyDetector
from typing import Optional
import pyspark.sql.functions as F
import plotly.express as px
from plotly.graph_objects import Figure
import pandas as pd


class ResidualAnomalyDetector(AnomalyDetector):
    """

    """
    @staticmethod
    def detect(ts: DataFrame, time_col_name: str, resid_col_name: str, ad_hoc: Optional[bool] = True, upper_limit: Optional[float] = None, lower_limit: Optional[float] = None) -> AnalysisReport:
        """
        Anomaly detection by residual distribution. It use 6-sigma rule now.
        :param ts:
        :param time_col_name:
        :param resid_col_name:
        :param ad_hoc:
        :param upper_limit:
        :param lower_limit:
        :return:
        """
        # todo: parameterize confidence interval
        # validate parameters #1: if ad_hoc is false, then lower_limit, upper_limit must be passed
        if not ad_hoc and (lower_limit is None or upper_limit is None):
            raise TypeError("If ad_hoc is false, then lower_limit, upper_limit must be passed")

        # ad-hoc distribution analysis, for obtaining mean, stddev, lower_limit(mean - 3*stddev), upper_limit(mean + 3*stddev)
        if ad_hoc:
            stat_dict = ts.agg(
                F.mean(resid_col_name).alias(
                    "mean"
                ),
                F.stddev(resid_col_name).alias(
                    "stddev"
                )
            ).first().asDict()

            # shadow parameters
            mean = stat_dict["mean"]
            stddev = stat_dict['stddev']

            lower_limit = mean - 3 * stddev
            upper_limit = mean + 3 * stddev

        # detect outliers (classify)
        condition = (F.col(resid_col_name) <= lower_limit) | (F.col(resid_col_name) >= upper_limit)
        anomaly_detected_df = ts.withColumn("is_anomaly_by_residual", F.when(condition, True).otherwise(False))  # checked null -> false, # todo: testcase..

        # figure plot
        plot = ResidualAnomalyDetector.__build_plot(ts, time_col_name, resid_col_name, lower_limit, upper_limit)

        # build report
        report = AnalysisReport()
        report.parameters["lower_limit"] = lower_limit
        report.parameters["upper_limit"] = upper_limit
        report.parameters["anomaly_detected_df"] = anomaly_detected_df
        report.plots['resid_plot'] = plot
        return report

    @staticmethod
    def __build_plot(ts: DataFrame, time_col_name: str, resid_col_name: str, lower_limit: float, upper_limit: float) -> Figure:
        ts_pd = ts.sort(time_col_name).select(time_col_name, resid_col_name).toPandas()
        ts_pd[time_col_name] = pd.to_datetime(ts_pd[time_col_name], unit='ms')
        fig = px.line(ts_pd, x=time_col_name, y=resid_col_name)
        fig.add_hline(y=upper_limit, line_dash="dot",
                      annotation_text="mean(resid) + stddev(resid) * 3",
                      annotation_position="top left")
        fig.add_hline(y=lower_limit, line_dash="dot",
                      annotation_text="mean(resid) - stddev(resid) * 3",
                      annotation_position="bottom left")
        return fig



"""

"""

from univariate.missing_value.handler.strategy import MissingValueHandleStrategy, MissingValueHandleStrategyType
from pyspark.sql import DataFrame
import pyspark.sql.functions as F
from univariate.analyzer import AnalysisReport
from univariate.analyzer import DescriptiveStatAnalyzer
from typing import Optional
from pyspark.sql.window import Window
from scipy.interpolate import CubicSpline
import numpy as np
from pyspark.sql.types import FloatType
from typing import Tuple

__all__ = ["CubicSplineInterpolation"]


class CubicSplineInterpolation(MissingValueHandleStrategy):
    """ """

    def handle(
            self, missed_ts: DataFrame, time_col_name: str, data_col_name, handled_col_name: str
    ) -> DataFrame:
        """

        :param missed_ts:
        :param time_col_name:
        :param data_col_name:
        :param handled_col_name:
        :return:
        """
        # todo: documentation
        xm, xscale, ym, yscale, poly = model_spline_interpolator(missed_ts, time_col_name, data_col_name)
        interpolate_cubic_spline_udf = F.udf(lambda x: (poly((x - xm) / xscale) * yscale + ym).item(), FloatType())
        filled_spline_df = missed_ts.withColumn(handled_col_name, F.when(F.col(data_col_name).isNull(),
            interpolate_cubic_spline_udf(F.col(time_col_name))).otherwise(F.col(data_col_name)))
        return filled_spline_df


def model_spline_interpolator(ts: DataFrame, time_col_name: str, data_col_name: str) -> Tuple[np.float64, np.float64, np.float64, np.float64, CubicSpline]:
    not_null_pd = ts.filter(F.col(data_col_name).isNotNull()).sort(time_col_name).toPandas()
    x = not_null_pd[time_col_name].values
    y = not_null_pd[data_col_name].values

    xm = np.mean(x)
    xscale = np.std(x)

    ym = np.mean(y)
    yscale = np.std(y)

    x = (x - xm) / xscale
    y = (y - ym) / yscale

    poly = CubicSpline(x, y)
    return xm, xscale, ym, yscale, poly



"""

"""

from pyspark.sql import DataFrame


class DuplicateReport:
    """

    """
    # todo: refactor generalization with analysis report, property pattern, report builder and so on
    def __init__(self, duplicate_df: DataFrame):
        self.duplicated = True if duplicate_df.count() > 0 else False
        self.duplicate_df = duplicate_df or None

    def __str__(self):
        pass



"""

"""

from abc import ABCMeta, abstractmethod
from pyspark.sql import DataFrame


class Validator(metaclass=ABCMeta):
    """
    Callable validator then ts is valid, before analyze
    """

    @abstractmethod
    def validate(self, ts: DataFrame, time_col_name: str):
        pass



from univariate.sampling.strategy.sampling_strategy import SamplingStrategy, SamplingStrategyType
# from univariate.sampling.strategy.fourier_method_sampling import FourierMethodSampling
from univariate.sampling.strategy.mean_downsampling import MeanDownsampling
from univariate.sampling.strategy.linear_upsampling import LinearUpsampling



"""

"""

from univariate.sampling.strategy import SamplingStrategy
from univariate.sampling.utils import find_start_timestamp_by_freq, construct_range_array
from pyspark.sql import DataFrame, SparkSession
from typing import Optional, List, Tuple
from pyspark.sql import functions as F
from pyspark.sql.types import LongType, IntegerType


class MeanDownsampling(SamplingStrategy):
    """

    """
    @staticmethod
    def sample(ts: DataFrame, time_col_name: str, data_col_name: str, period: int, sampling_period: int, by_freq: bool, count_flag_cols: List[str], display_point: Optional[str] = "first") -> DataFrame:
        """

        :param ts:
        :param time_col_name:
        :param data_col_name:
        :param period:
        :param sampling_period:
        :param by_freq:
        :param count_flag_cols:
        :param display_point:
        :return:
        """
        logger = SparkSession.getActiveSession()._jvm.org.apache.log4j.LogManager.getLogger(
            __name__
        )
        # todo: validate display_point parameter
        # first range's start timestamp
        if by_freq:
            first_timestamp = ts.sort(time_col_name).first().asDict()[time_col_name]
            first_start = find_start_timestamp_by_freq(first_timestamp, sampling_period)
            last_timestamp = ts.sort(time_col_name).tail(1)[0].asDict()[time_col_name]
            last_start = find_start_timestamp_by_freq(last_timestamp, sampling_period)

            start_series, end_series = construct_range_array(sampling_period, first_start, last_start, True)
            logger.info(f"start_series {start_series}")
            logger.info(f"end_series {end_series}")
            # partition col by intervals
            partitioned_df = ts.withColumn("_range_start", partitionize_udf(start_series, end_series)(F.col(time_col_name)))

        # aggregation
        agg_cols = [F.mean(data_col_name).alias(data_col_name),  F.count("*").alias("count")]
        select_cols = ["_range_start", data_col_name, "count"]
        cnt_cond = lambda cond: F.sum(F.when(cond, 1).otherwise(0))
        if len(count_flag_cols) > 0:
            agg_cols.extend([cnt_cond(F.col(col)).alias(col + "_count") for col in count_flag_cols])
            select_cols.extend([col + "_count" for col in count_flag_cols])


        agg_df = partitioned_df.groupBy("_range_start").agg(*agg_cols).select(*select_cols)

        # display_name
        if display_point == "first":
            sampled_df = agg_df.withColumnRenamed("_range_start", time_col_name)
        elif display_point == "middle":
            sampled_df = agg_df.withColumn(time_col_name, F.col("_range_start") + F.lit(sampling_period // 2)).drop("_range_start")
        elif display_point == "last":
            sampled_df = agg_df.withColumn(time_col_name, F.col("_range_start") + F.lit(sampling_period)).drop("_range_start")

        # sort
        return sampled_df.sort(time_col_name)

'''
def calc_range_start(timestamp: int, first_start: int, sampling_period: int) -> int:
    """

    :param timestamp:
    :param first_start:
    :param sampling_period:
    :return:
    """
    return ((timestamp - first_start) // sampling_period) * sampling_period + first_start
'''


def partitionize(timestamp: int, start_series: List[int], end_series: List[int]) -> int:
    """

    :param timestamp:
    :param start_series:
    :param end_series:
    :return:
    """
    for i in range(len(start_series)):
        if start_series[i] <= timestamp and end_series[i] > timestamp:
            return start_series[i]


def partitionize_udf(start_series: List[int], end_series: List[int]):
    return F.udf(lambda t: partitionize(t, start_series, end_series))



"""
    DTO for reporting analysis result
"""


class AnalysisReport:
    """
    DTO for reporting analysis result
    """

    def __init__(self):
        """
        # todo: temp report, spec latter
        """
        self.metrics = dict()
        self.parameters = dict()
        self.plots = dict()



"""

"""

from univariate.analyzer import Analyzer
from univariate.analyzer import AnalysisReport
from pyspark.sql import DataFrame
import plotly.express as px
from univariate.sampling.utils import freq_to_period_map
from statsmodels.tsa.seasonal import STL
import pandas as pd
from pyspark.sql import SparkSession
from pyspark.sql import functions as F


class DecompositionAnalyzer(Analyzer):
    """
    Decompose Time series
    """
    def __init__(self):
        """

        """
        # todo: decide decomposition strategy
        pass

    def analyze(self, ts: DataFrame, time_col_name: str, data_col_name: str, seasonal_freq: str, ts_period: int) -> AnalysisReport:
        """

        :param ts:
        :param time_col_name:
        :param data_col_name:
        :param seasonal_freq:
        :param ts_period: milli seconds
        :return:
        """
        period = self.__calc_period(seasonal_freq, ts_period)
        ts_pd = ts.sort(time_col_name).select(time_col_name, data_col_name).toPandas()
        ts_pd[time_col_name] = pd.to_datetime(ts_pd[time_col_name], unit='ms')  # todo: ms constraint?
        stl = STL(ts_pd.set_index(time_col_name)[data_col_name], period=period).fit()  # todo: strategies for ts decompositions  #todo: multi seasonal

        # pandas datetime drop timestamp's millisec, so recover with original timestamp from ts
        timestamp_series = ts.select(time_col_name).sort(time_col_name).toPandas()[time_col_name]
        decomposed_pd = pd.concat([stl.trend, stl.seasonal, stl.resid], axis=1).reset_index(drop=True)
        decomposed_pd[time_col_name] = timestamp_series
        decomposed_df = SparkSession.getActiveSession().createDataFrame(decomposed_pd)
        decomposed_df = ts.join(decomposed_df, time_col_name, "left").sort(time_col_name)

        report = AnalysisReport()
        report.parameters['decomposed_df'] = decomposed_df
        report.plots["decomposed"] = stl
        report.plots["observed"] = px.line(stl.observed)
        report.plots["seasonal"] = px.line(stl.seasonal)
        report.plots["trend"] = px.line(stl.trend)
        report.plots["resid"] = px.line(stl.resid)

        return report


    def __calc_period(self, seasonal_freq: str, ts_period: int):
        """

        :param seasonal_freq:
        :param ts_period:
        :return:
        """
        if seasonal_freq not in freq_to_period_map.keys():
            raise ValueError(f"seasonal_freq {seasonal_freq} is not supported")
        if (freq_to_period_map[seasonal_freq] / 2) < ts_period:
            raise ValueError(f"Seasonal component need at least 2 observed points. So ts_period should be less than half of seasonal period.")

        return int(round(freq_to_period_map[seasonal_freq] / ts_period))
